import asyncio
import logging
from zof.event import load_event

LOGGER = logging.getLogger(__package__)


class Protocol(asyncio.SubprocessProtocol):
    """Implements an asyncio Protocol for parsing data received from oftr.
    """

    def __init__(self, post_event):
        self.post_event = post_event
        self.buf = b''
        self.exit_future = asyncio.Future()

    def pipe_data_received(self, fd, data):
        LOGGER.debug('zof.Protocol.pipe_data_received: %d bytes, fd=%d',
                     len(data), fd)
        begin = 0
        offset = len(self.buf)
        self.buf += data
        while True:
            offset = self.buf.find(b'\x00', offset)
            if offset < 0:
                self.buf = self.buf[begin:]
                return
            if begin != offset:
                self.post_event(load_event(self.buf[begin:offset]))
            offset += 1
            begin = offset

    def pipe_connection_lost(self, fd, exc):
        if exc is not None:
            LOGGER.warning('zof.Protocol.pipe_connection_lost: fd=%d, exc=%r',
                           fd, exc)

    def process_exited(self):
        LOGGER.debug('zof.Protocol.process_exited')
        self.post_event(load_event(b''))
        self.exit_future.set_result(0)



import asyncio
import json
import os
import re
import signal
import tempfile
import zof

#       zof <---------+
#         |           |
#      exabgp         |
#         |           |
#     bgp_tunnel -----+

APP = zof.Application('bgp')


class BGPNeighbor:
    """Represents a peer connection to a BGP neighbor."""

    EXABGP_CONF = b'''
        neighbor %(peer_address)a {
            router-id %(router_id)a;
            local-as %(local_as)a;
            local-address %(local_address)a;
            peer-as %(peer_as)a;
            api {
                processes [tunnel];
                neighbor-changes;
                receive { 
                    open;
                    update;
                    parsed;
                }
            }
        }
    '''

    def __init__(self, local_as, local_address, router_id, peer_as, peer_address):
        self.local_as = local_as
        self.local_address = local_address
        self.router_id = router_id
        self.peer_as = peer_as
        self.peer_address = peer_address

    def conf(self):
        return self.EXABGP_CONF % {
            b'local_as': self.local_as,
            b'local_address': self.local_address,
            b'router_id': self.router_id,
            b'peer_as': self.peer_as,
            b'peer_address': self.peer_address
        }


class BGPServer:
    """BGP Server implemented using exabgp."""

    EXABGP_CONF = b'''
        process tunnel {
          run /usr/bin/nc -U %(server_path)a;
          encoder json;
        }

        %(neighbor_conf)b
    '''

    EXABGP_ENV = b'''
        [exabgp.api]
        cli = false
        [exabgp.tcp]
        bind = %(tcp_bind)a
        port = %(tcp_port)d  
    '''

    def __init__(self):
        self.server_path = os.path.join(os.path.dirname(__file__), 'zof_bgp.sock')
        self.server = None
        self.proc = None
        self.exabgp_conf = tempfile.NamedTemporaryFile()
        self.exabgp_env = tempfile.NamedTemporaryFile()
        self.neighbors = []
        APP.logger.info('BGPServer path=%s', self.server_path)

    def __del__(self):
        self.exabgp_conf.close()
        self.exabgp_env.close()

    def add_neighbor(self, local_as, local_address, router_id, peer_as, peer_address):
        """Add BGP neighbor."""
        neighbor = BGPNeighbor(local_as, local_address, router_id, peer_as, peer_address)
        self.neighbors.append(neighbor)
        self._write_conf()
        self._reload_conf()
        return neighbor

    async def start(self):
        """Start BGP server."""
        assert self.server is None
        self.server = await asyncio.start_unix_server(self._accept, path=self.server_path)
        await self._run_exabgp()

    async def stop(self):
        """Stop BGP server."""
        if self.server is None:
            return
        self.server.close()
        try:
            self.proc.terminate()
        except ProcessLookupError:
            pass

    async def _accept(self, reader, _writer):
        """Read from connection to exabgp tunnel and post BGP events."""
        while True:
            line = await reader.readline()
            if not line:
                break
            message = self._decode_message(line)
            if message:
                zof.post_event(message)

    def _decode_message(self, line):
        """Decode JSON message from exabgp."""
        try:
            line = _fix_exabgp_line(line)
            message = json.loads(line)
            assert 'event' not in message
            message['event'] = 'BGP'
            APP.logger.info(repr(message))
            return message         
        except json.decoder.JSONDecodeError:
            APP.logger.error("JSON error: %s", line)
            return None

    async def _run_exabgp(self):
        """Run exabgp with tunnel subprocess."""
        self._write_conf()
        self._write_env()
        args = ['exabgp', '-e', self.exabgp_env.name, self.exabgp_conf.name]
        self.proc = await asyncio.create_subprocess_exec(*args)

    def _write_conf(self):
        """Construct temporary file with exabgp config."""
        neighbor_conf = b'\n'.join(neigh.conf() for neigh in self.neighbors)
        self.exabgp_conf.truncate(0)
        self.exabgp_conf.write(self.EXABGP_CONF % {
            b'server_path': self.server_path,
            b'neighbor_conf': neighbor_conf})
        self.exabgp_conf.flush()

    def _write_env(self, tcp_bind='127.0.0.1', tcp_port=1790):
        """Construct temporary file with exabgp envionment variables."""
        self.exabgp_env.write(self.EXABGP_ENV % {
            b'tcp_bind': tcp_bind,
            b'tcp_port': tcp_port})
        self.exabgp_env.flush()

    def _reload_conf(self):
        """Signal exabgp to reload its config."""
        APP.logger.info('sending signal...')
        #self.proc.send_signal(signal.SIGUSR1)


FIX_REGEX = re.compile(br'"reason": (peer reset, message \[[^\]]+\] error\[[^\]]+\])')

def _fix_exabgp_line(line):
    return FIX_REGEX.sub(br'"reason": "\1"', line)


@APP.event('start')
async def start(_):
    APP.server = BGPServer()
    await APP.server.start()
    neigh = APP.server.add_neighbor(1, '127.0.0.1', '1.1.1.1', 2, '127.0.0.1')


@APP.event('stop')
async def stop(_):
    await APP.server.stop()


if __name__ == '__main__':
    zof.run()



import inspect


class CallbackInfo:
    """Concrete class encapsulating information about a callback function.

    This class supports callbacks that may be registered by a decorator. The
    callback can be normal function or a method that is bound to an object 
    later. A callback may be an asyncio coroutine function; the client is
    responsible for calling it correctly.

    Example:

        class Foo:
            def cb(self, arg):
                ...

        info = CallbackInfo(Foo.cb)
        assert info.arg_count == 1

        func = info.bind(Foo())
        func(1)
    """

    def __init__(self, callback):
        """Inspect callback function."""
        # Track type of original callback.
        func = callback
        orig_type = type(func).__name__
        assert orig_type in {'method', 'function', 'staticmethod', 'classmethod'}

        # Dereference staticmethod and classmethod wrappers. These appear when
        # invoked inside a decorator.
        if isinstance(func, (staticmethod, classmethod)):
            func = func.__func__

        # Obtain function's signature and qualified name.
        assert callable(func)
        sig = inspect.signature(func)
        qualname = func.__qualname__

        # A method will have a __self__ attribute that can tell us the class
        # name. For other types, parse the end of the qualified name.
        class_ = None
        if orig_type == 'method':
            cls = func.__self__.__class__
            if cls is type:  # class method
                class_ = func.__self__.__name__
            else:
                class_ = cls.__name__
        elif '.' in qualname:
            lhs, _ = qualname.split('.')[-2:]
            if lhs != '<locals>':
                class_ = lhs

        # Obtain list of parameters and check if instance is required.
        params = list(sig.parameters.values())
        instance_required = _instance_required(orig_type, class_, params)
        class_required = _class_required(orig_type, class_, params)

        # Count the number of required positional arguments.
        args_required = _args_required(params)
        if instance_required or class_required:
            args_required -= 1

        # If staticmethod passed in, dereference it here.
        if orig_type == 'staticmethod':
            callback = func

        self.name = '%s%s' % (qualname, sig)
        self.callback = callback
        self.class_ = class_
        self.args_required = args_required
        self.instance_required = instance_required
        self.class_required = class_required

    def bind(self, instance=None):
        """Return callback bound to instance."""
        callback = self.callback
        # Handle class binding and instance binding.
        if self.class_required or self.instance_required:
            if instance is None:
                raise ValueError('Callback %s requires an instance of class %s' % (self.name, self.class_))
            cls = instance.__class__
            if cls.__name__ != self.class_:
                raise ValueError('Callback %s requires an instance of class %s (not %s)' % (self.name, self.class_, cls.__name__))
            return callback.__get__(instance, cls)
        return callback

    @property
    def bind_required(self):
        """Return true if callback must be bound to an instance."""
        return self.class_required or self.instance_required


def _args_required(params):
    """Return count of positional arguments needed."""
    return len([param for param in params if param.default == inspect.Parameter.empty])


def _instance_required(type_, class_, params):
    """Return true if method requires `self` instance."""
    if not params or class_ is None or type_ != 'function':
        return False
    # Check that the first parameter is named "self" to distinguish an
    # unbound method from a dereferenced staticmethod.
    return params[0].name == 'self' and params[0].default == inspect.Parameter.empty


def _class_required(type_, class_, params):
    """Return true if method requires a `cls` instance."""
    if not params or class_ is None:
        return False
    return type_ == 'classmethod'



import unittest
import asyncio
from zof.callbackinfo import CallbackInfo


class CallbackInfoTestCase(unittest.TestCase):

    def test_inspect_decorated(self):
        items = [_tuple(item) for item in DECORATED]
        expected = [
            ('fa(a)', 1, None, False, False), 
            ('fb(a, b=3)', 1, None, False, False), 
            ('afa(a)', 1, None, False, False), 
            ('afb(a, b=3)', 1, None, False, False), 
            ('local_func.<locals>.fc(a)', 1, None, False, False), 
            ('local_func.<locals>.Foo.fd(self, a)', 1, 'Foo', True, False), 
            ('Foo.fe(self, a)', 1, 'Foo', True, False), 
            ('Foo.ff(a, b=1)', 1, 'Foo', False, False), 
            ('Foo.fg(a)', 1, 'Foo', False, False), 
            ('Foo.fh(cls, a)', 1, 'Foo', False, True), 
            ('Foo.afe(self, a)', 1, 'Foo', True, False), 
            ('Foo.aff(a, b=1)', 1, 'Foo', False, False), 
            ('Foo.afg(a)', 1, 'Foo', False, False), 
            ('Foo.afh(cls, a)', 1, 'Foo', False, True)
        ]
        self.assertEqual(items, expected)

    def test_inspect_direct(self):
        items = [_tuple(CallbackInfo(func)) for func in DIRECT]
        expected = [
            ('fa(a)', 1, None, False, False), 
            ('fb(a, b=3)', 1, None, False, False), 
            ('afa(a)', 1, None, False, False), 
            ('afb(a, b=3)', 1, None, False, False), 
            ('Foo.fe(self, a)', 1, 'Foo', True, False), 
            ('Foo.ff(a, b=1)', 1, 'Foo', False, False), 
            ('Foo.fg(a)', 1, 'Foo', False, False), 
            ('Foo.fh(a)', 1, 'Foo', False, False),
            ('Foo.afe(self, a)', 1, 'Foo', True, False), 
            ('Foo.aff(a, b=1)', 1, 'Foo', False, False), 
            ('Foo.afg(a)', 1, 'Foo', False, False), 
            ('Foo.afh(a)', 1, 'Foo', False, False)
        ]
        self.assertEqual(items, expected)

    def test_bind_decorated(self):
        instance = Foo()
        results = [_bindcall1(item, instance) for item in DECORATED]
        expected = ['fa', 'fb', 'afa', 'afb', 'fc', 'fd', 'fe', 'ff', 'fg', 'fh', 'afe', 'aff', 'afg', 'afh']
        self.assertEqual(results, expected)

    def test_bind_direct(self):
        items = [CallbackInfo(func) for func in DIRECT]
        instance = Foo()
        results = [_bindcall1(item, instance) for item in items]
        expected = ['fa', 'fb', 'afa', 'afb', 'fe', 'ff', 'fg', 'fh', 'afe', 'aff', 'afg', 'afh']
        self.assertEqual(results, expected)



def _tuple(info):
    return (info.name, info.args_required, info.class_, info.instance_required, info.class_required)

def _bindcall1(info, obj):
    func = info.bind(obj)
    if asyncio.iscoroutinefunction(func):
        return _run(func(1))
    return func(1)

def _run(coro):
    loop = asyncio.new_event_loop()
    result = loop.run_until_complete(coro)
    loop.close()
    return result


DECORATED = []

def inspectx(func):
    DECORATED.append(CallbackInfo(func))
    return func


# Test functions.

@inspectx
def fa(a):
    return 'fa'

@inspectx
def fb(a, b=3):
    return 'fb'

@inspectx
async def afa(a):
    return 'afa'

@inspectx
async def afb(a, b=3):
    return 'afb'

# Test locally defined functions.

def local_func():
    @inspectx
    def fc(a):
        return 'fc'

    class Foo:
        @inspectx
        def fd(self, a):
            return 'fd'

local_func()

# Test class methods.

class Foo:
    # Single arg method
    @inspectx
    def fe(self, a):
        return 'fe'

    # Single arg method, with default arg
    @inspectx
    def ff(a, b=1):
        return 'ff'

    # Static method
    @inspectx
    @staticmethod
    def fg(a):
        return 'fg'

    # Class method
    @inspectx
    @classmethod
    def fh(cls, a):
        return 'fh'

    # Single arg method (async)
    @inspectx
    async def afe(self, a):
        return 'afe'

    # Single arg method, with default arg (async)
    @inspectx
    async def aff(a, b=1):
        return 'aff'

    # Static method (async)
    @inspectx
    @staticmethod
    async def afg(a):
        return 'afg'

    # Class method (async)
    @inspectx
    @classmethod
    async def afh(cls, a):
        return 'afh'


DIRECT = (fa, fb, afa, afb, Foo.fe, Foo.ff, Foo.fg, Foo.fh, Foo.afe, Foo.aff, Foo.afg, Foo.afh)



import zof


APP = zof.Application('layer2mt')


@APP.bind()
class Layer2_MT:

    @APP.message('channel_up')
    @APP.message('channel_down')
    @staticmethod
    def channel_event(event):
        APP.logger.info(event)


if __name__ == '__main__':
    zof.run()
